const Controller = require('./controller');
const PubSub = require('pubsub-js');
const { menuEvent, requester, receiver } = require('../main/context');
const { scattering, matrix } = require('../app-state');
const { calculateKiKf } = require('../utils/calculations');
const THREE = require('three');

class ScatteringController extends Controller {
    constructor(scatteringPanel, anglesPanel, scattering3D) {
        super();

        this._scatteringPanel = scatteringPanel;
        this._anglesPanel = anglesPanel;
        this._scattering3D = scattering3D;

        // Action handlers.
        // Update only the panel.
        PubSub.subscribe('update-panel', (msg, value) => { this.updateQKiKf(value.qPoints, value.kiPoints, value.kfPoints, value.ss); });

        // Update only the 3D view.
        PubSub.subscribe('update-3D', (msg, value) => { this.update3D(value.qPoints, value.kiPoints, value.kfPoints, value.ss); });
        PubSub.subscribe('update-3D-local', (msg, value) => { this.update3DLocal(value.qPoints, value.kiPoints, value.kfPoints); });

        PubSub.subscribe('fix-scattering', (msg, value) => { this.fixScattering(value); });
        PubSub.subscribe('pause-scattering', (msg, value) => { this.pause(value); });

        // Events handlers.
        PubSub.subscribe('change-scattering', () => { this.scatteringChange(); });
        PubSub.subscribe('change-plane', () => { this.projectArrows(); });
        PubSub.subscribe('change-sample', () => { this.projectArrows(); });

        receiver.on('update-angles', (msg, data) => {
            scattering.a1 = data.a1;
            scattering.a2 = data.a2;
            scattering.a3 = data.a3;
            scattering.a3p = data.a3p;
            scattering.a4 = data.a4;
            scattering.a5 = data.a5;
            scattering.a6 = data.a6;
            this.updateAngles();
        });

        menuEvent.on('change-mode', (event, value) => {
            this._scatteringPanel.setMode(value);
        });

        // Initializations.
        this.init();
    }

    init() {
        // Set the free mode by default.
        this._scatteringPanel.setMode('free');

        // Recalculate the points.
        // The Ki and Kf points are not coming from Nomad and need to be recalculated.
        this.scatteringChange();
    }
   
    /**
     * Update scattering in response to scattering model update.
     */
    scatteringChange() {

        this.updateScattering();
        this.updateAngles();
        this.updateScatteringPanel();
        this._scattering3D.setScattering(scattering);
    }

    /**
     * Recalculate the Ki and Kf points.
     */
    updateScattering() {

        // Get q in R frame.
        let q = new THREE.Vector3(scattering.qh, scattering.qk, scattering.ql).applyMatrix3(matrix.bMatrix);
        
        // Calculate the new points.
        let result = calculateKiKf(q, scattering.ki, scattering.kf, scattering.ss, matrix.uMatrix_1, matrix.uMatrix);

        // Set the calculated points.
        scattering.kiPoints = result.kiPoints;
        scattering.kfPoints = result.kfPoints;
    }

    /**
     * Projects arrows on plane when it moves.
     */
    projectArrows() {
        this._scattering3D.project();
    }

    /**
     * Set ki, kf or q to fixed length.
     */
    fixScattering(value) {
        requester.request('fixScattering', value, (data) => {
            scattering.update(data);
        });
    }

    /**
     * Update angles panel.
     */
    updateAngles() {
        this._anglesPanel.setInputs(scattering);
    }

    /**
     * Update the panel input using model values.
     */
    updateScatteringPanel() {
        this._scatteringPanel.setInputs(scattering);
    }

    //// Update Q, ki and kf using the reciprocal view ////
    /// Update views locally without informing the server.
    update3DLocal(qVec, kiPoints, kfPoints) {

        scattering.qh = qVec.x;
        scattering.qk = qVec.y;
        scattering.ql = qVec.z;
    }

    /// Update views and inform the server.
    update3D(qVec, kiPoints, kfPoints, ss) {
        requester.request('setPoints', {
            qPoints: qVec, kiPoints, kfPoints, ss
        }, (data) => {
            scattering.update(data);
            this.updateScatteringPanel();
            // Do not update angles as it is done asynchronously.
        });
    }
    
    updateQKiKf(qVec, kiPoints, kfPoints, ss) {
        requester.request('setPoints', {
            qPoints: qVec, kiPoints, kfPoints, ss
        }, (data) => {
            scattering.update(data);
            this._scattering3D.setScattering(scattering);
            this.updateScatteringPanel();
            // Do not update angles as it is done asynchronously.
        });
    }
        
    pause(value) {
        scattering.paused = value;
        matrix.paused = value;
    }
    
}

module.exports = ScatteringController;