const THREE = require('three');

class CameraMove {

    constructor(camera) {
        this._camera = camera;

        this._positions = new THREE.CatmullRomCurve3();
        this._targets = new THREE.CatmullRomCurve3();

        this._index = 0;
        this._enableMove = false;

        this._listeners = {};

        let self = this;

        function updateLoop() {
            self.move();
            setTimeout(updateLoop, 10);
        }

        updateLoop();
    }

    set enableMove(value) { this._enableMove = value; }

    /**
     * Register a listener.
     */
    on(event, listener) {
        if(!this._listeners[event]) {
            this._listeners[event] = [listener];
        } else {
            this._listeners[event].push(listener);
        }
    }

    /**
     * Dispatch an event to all listeners.
     */
    dispatch(event, data) {
        if(this._listeners[event]) {
            for(let listener of this._listeners[event]) {
                listener(data);
            }
        }
    }

    /**
     * Add a waypoint to camera trajectory.
     */
    addWaypoint(point, target) {
        let points = this._positions.points;
        let targets = this._targets.points;

        points.push(point);
        targets.push(target);

        this._positions = new THREE.CatmullRomCurve3(points);
        this._targets = new THREE.CatmullRomCurve3(targets);
    }

    /**
     * Clear the camera trajectory.
     */
    clear() {
        this._positions = new THREE.CatmullRomCurve3();
        this._targets = new THREE.CatmullRomCurve3();
        this._index = 0;
        this._enableMove = false;
    }

    /**
     * Move the camera.
     */
    move() {
        if(this._enableMove) {
            if(this._index >= 1) {
                this.dispatch('moveend');
                this._enableMove = false;
                return;
            }

            this._index += 0.01;

            var camPos = this._positions.getPoint(this._index);
            var camTarget = this._targets.getPoint(this._index);

            this._camera.position.copy(camPos);
            this._camera.lookAt(camTarget);
        }
    }
}

module.exports = CameraMove;