const Model = require('./model');
const vEXPController = require('../vexp-controller');
const xtal = require('./xtal/xtal');

class Atom {
    constructor(name, x, y, z) {
        this._name = name;
        this._x = x;
        this._y = y;
        this._z = z;
    }

    // Getters.
    get name() { return this._name; }
    get x() { return this._x; }
    get y() { return this._y; }
    get z() { return this._z; }

    // Setters.
    set name(value) { this._name = value; }
    set x(value) { this._x = value; }
    set y(value) { this._y = value; }
    set z(value) { this._z = value; }

    /**
     * Verify that 2 atoms are equals.
     */
    equals(atom) {
        return this._name === atom.name && 
               this._x === atom.x &&
               this._y === atom.y &&
               this._z === atom.z;
    }

    /**
     * Cast this instance into a basic JS object. Used
     * for configuration saving.
     */
    toObject() {
        return {
            name: this._name,
            x: this._x,
            y: this._y,
            z: this._z
        };
    }
}

class AtomsModel extends Model {
    constructor() {
        super();

        this._atoms = [];
        this._atomnames = [];
        this._atompos = [];
    }

    // Getters.
    get atoms() { return this._atoms; }
    get atompos() { return this._atompos; }
    get atomnames() { return this._atomnames; }

    /**
     * Update atom list in vEXPController.
     */
    updateVEXPController() {
        if(xtal.scatteringLengths === null)
            return;

        this._atomnames = [];
        this._atompos = [];

        for(let atom of this._atoms) {
            this._atomnames.push(atom.name);
            this._atompos.push([ atom.x, atom.y, atom.z ]);
        }

        vEXPController.setAtomsFromUI(this._atomnames, this._atompos);
    }

    /**
     * Find an atom by its name and position.
     */
    _getAtom(name, x, y, z) {
        let refAtom = new Atom(name, x, y, z);
        for(let atom of this._atoms) {
            if(atom.equals(refAtom)) {
                return atom;
            }
        }

        return null;
    }

    /**
     * Add an atom to the list.
     */
    addAtom(name, x, y, z) {
        let atom = new Atom(name, x, y, z);
        this._atoms.push(atom);
        this.updateVEXPController();
        return atom;
    }

    /**
     * Delete an atom from the list.
     */
    delAtom(index) {
        this._atoms.splice(index, 1);
        this.updateVEXPController();

        return null;
    }

    /**
     * Remove all atoms from the list.
     */
    clearAtoms() {
        this._atoms = [];
        this.updateVEXPController();
    }

    /**
     * Update an atom in the list.
     */
    updateAtom(index, newAtom) {
        let atom = this._atoms[index];
        atom.name = newAtom.name;
        atom.x = newAtom.x;
        atom.y = newAtom.y;
        atom.z = newAtom.z;

        this.updateVEXPController();
    }

    toObject() {
        let atomsCopy = [];

        for(let at of this._atoms) {
            atomsCopy.push(at.toObject());
        }

        return {
            atoms: atomsCopy,
            atompos: this._atompos.slice(),
            atomnames: this._atomnames.slice()
        }
    }

    /**
     * Save the model data into a config object (the saver object).
     */
    save(saver) {
        saver.atoms = this.toObject();
    }

    /**
     * Restore the model state saved in the saver object (the configuration object).
     */
    restore(saver) {
        let atoms = saver.atoms;
        
        this._atoms = [];

        for(let at of atoms.atoms) {
            this._atoms.push(new Atom(at.name, at.x, at.y, at.z));
        }

        this.updateVEXPController();
    }
}

module.exports = AtomsModel;