const THREE = require('three');
const Angle = require('../../tas/angle');
const TASFormulas = require('../../tas/tas-formulas');

var Mode = {
    ABC: "abc",
    AB: "ab",
    AC: "ac",
    BC: "bc",
    NONE: "none"
}
Mode.fromString = new function (string) {
    for (let t in Mode) {
        val = Mode[t];
        if (string == val.toString())
            return val;
    }
    return Mode.NONE;
}

class Sample {

    constructor(name, description, type, a,b,c, alpha,beta,gamma, orientation1,orientation2) {
        // ---Name and description of this sample---
        this.name = name;
        this.description = description;
        this.type = type; // Must be here, as setA() setB() setC() below depend on it...

        this.setA = a;
        this.setB = b;
        this.setC = c;
        this.setAlpha = new Angle(alpha/180*Math.PI);
        this.setBeta = new Angle(beta/180*Math.PI);
        this.setGamma = new Angle(gamma/180*Math.PI);
  
        //These are used for hkl1 and hkl2
        this.orientation1 = orientation1;
        this.orientation2 = orientation2;
 
        this.u_matrix = new THREE.Matrix3();
        this.b_matrix = new THREE.Matrix3();
        this.ub_matrix = new THREE.Matrix3();
    }

    equals(o) {
        let result;
        if (this === o) {
            return true;
        }

        if (!(o instanceof Sample)) {
            return false;
        }

        result = this.name.localeCompare(o.name) == 0 && this.description.localeCompare(o.description) == 0 &&
            Math.abs(this.a - o.a) < TASFormulas.CMP_EPSILON &&
            Math.abs(this.b - o.b) < TASFormulas.CMP_EPSILON &&
            Math.abs(this.c - o.c) < TASFormulas.CMP_EPSILON &&
            Math.abs(this.alpha - o.alpha) < TASFormulas.CMP_EPSILON &&
            Math.abs(this.beta - o.beta) < TASFormulas.CMP_EPSILON &&
            Math.abs(this.gamma - o.gamma) < TASFormulas.CMP_EPSILON
        ;
        return result;
    }

    setDirectParameters(a, b, c, alpha, beta, gamma) {
        this.a = a;
        this.b = b;
        this.c = c;
        this.alpha = alpha;
        this.beta = beta;
        this.gamma = gamma;
        this.calculateReciprocals();
    }


    setReciprocalParameters(a_rec, b_rec, c_rec, alpha_rec, beta_rec, gamma_rec) {
        this.a_rec = a_rec;
        this.b_rec = b_rec;
        this.c_rec = c_rec;
        this.alpha_rec = alpha_rec;
        this.beta_rec = beta_rec;
        this.gamma_rec = gamma_rec;
        calculateDirectsUsingReciprocal();
    }

    calculateReciprocals() {
        let changed = false;

        if (this.a_valid && this.b_valid && this.c_valid && this.beta_valid && this.alpha_valid && this.gamma_valid) {
            // calcule du volume et + controle du signe de 'a', 'b' et 'c' par

            this.vol_real = TASFormulas.volume_of_unit_cell(this.a, this.b, this.c, this.alpha.getValue, this.beta.getValue, this.gamma.getValue);
            this.setVol_real = this.vol_real;
            this.setVol_rec = 1.0 / this.vol_real;
            if (this.vol_real_valid) {
                changed = true;
            }

            // Formula 4.5
            let newa_rec = TASFormulas.direct_from_to_rec_cell_lattice(this.b, this.c, this.alpha.getValue, this.vol_real);
            this.setA_rec(newa_rec);
            if (this.a_rec_valid) {
                changed = true;
            }

            // Formula 4.6
            let newb_rec = TASFormulas.direct_from_to_rec_cell_lattice(this.a, this.c, this.beta.getValue, this.vol_real);
            this.setB_rec(newb_rec);
            if (this.b_rec_valid) {
                changed = true;
            }

            // Formula 4.7
            let newc_rec = TASFormulas.direct_from_to_rec_cell_lattice(this.a, this.b, this.gamma.getValue, this.vol_real);
            this.setC_rec(newc_rec);
            if (this.c_rec_valid) {
                changed = true;
            }

            // Formula 4.8
            // angles must be ]0,180[
            let newalpha_rec = new Angle(TASFormulas.direct_from_to_rec_cell_angle(this.beta.getValue, this.gamma.getValue,
                this.alpha.getValue));

            this.setAlpha_rec(newalpha_rec);
            if (this.alpha_rec_valid) {
                changed = true;
            }

            // Formula 4.9
            let newbeta_rec = new Angle(TASFormulas.direct_from_to_rec_cell_angle(this.alpha.getValue, this.gamma.getValue,
                this.beta.getValue));

            this.setBeta_rec(newbeta_rec);
            if (this.beta_rec_valid) {
                changed = true;
            }

            // Formula 4.10
            let newgamma_rec = new Angle(TASFormulas.direct_from_to_rec_cell_angle(this.alpha.getValue, this.beta.getValue,
                this.gamma.getValue));

            this.setGamma_rec(newgamma_rec);
            if (this.gamma_rec_valid) {
                changed = true;
            }

        } else {
            this.a_rec_valid = false;
            this.b_rec_valid = false;
            this.c_rec_valid = false;

            this.alpha_rec_valid = false;
            this.beta_rec_valid = false;
            this.gamma_rec_valid = false;

            this.vol_real_valid = false;
            this.vol_rec_valid = false;
        }
        return changed;
    }
    
    calculateDirectsUsingReciprocal() {
        let changed = false;

        if (this.a_rec_valid && this.b_rec_valid && this.c_rec_valid && this.beta_rec_valid && this.alpha_rec_valid &&
            this.gamma_rec_valid) {
            // calul du volume reciproque
            this.vol_rec = TASFormulas.volume_of_unit_cell(this.a_rec, this.b_rec, this.c_rec, this.alpha_rec.getValue,
                this.beta_rec.getValue, this.gamma_rec.getValue);

            // nouvelle valeur de a
            let newa = TASFormulas.direct_from_to_rec_cell_lattice(this.b_rec, this.c_rec, this.alpha_rec.getValue,
                this.vol_rec);
            // affectation
            setA(newa);

            // nouvelle valeur de b
            let newb = TASFormulas.direct_from_to_rec_cell_lattice(this.a_rec, this.c_rec, this.beta_rec.getValue,
                this.vol_rec);
            // affectation
            setB(newb);

            // nouvelle valeur de c
            let newc = TASFormulas.direct_from_to_rec_cell_lattice(this.a_rec, this.b_rec, this.gamma_rec.getValue,
                this.vol_rec);
            // affectation
            setC(newc);

            // nouvelle valeur de alpha
            let newalpha = new Angle(TASFormulas.direct_from_to_rec_cell_angle(this.beta_rec.getValue,
                this.gamma_rec.getValue, this.alpha_rec.getValue));
            // affectation
            setAlpha(newalpha);

            // nouvelle valeur de beta
            let newbeta = new Angle(TASFormulas.direct_from_to_rec_cell_angle(this.alpha_rec.getValue,
                this.gamma_rec.getValue, this.beta_rec.getValue));
            // affectation
            setBeta(newbeta);

            let newgamma = new Angle(TASFormulas.direct_from_to_rec_cell_angle(this.alpha_rec.getValue,
                this.beta_rec.getValue, this.gamma_rec.getValue));
            setGamma(newgamma);
            changed = true;

            // Used new computed values for volume
            vol_real = TASFormulas.volume_of_unit_cell(this.a, this.b, this.c, this.alpha.getValue,
                this.beta.getValue, this.gamma.getValue);

            this.setVol_real(vol_real);
            this.setVol_rec(1 / vol_real);

        } else {
            a_valid = false;
            b_valid = false;
            c_valid = false;

            alpha_valid = false;
            beta_valid = false;
            gamma_valid = false;

            vol_real_valid = false;
            vol_rec_valid = false;

        }

        return changed;
    }

    /**
     * Calculate the UB Matrix of the source in its current orientation.
     * "COMPUTATION OF THE TRANSFORMATION MATRIX B AND ORIENTATION MATRIX S"
     *
     * @author unknown - 11-05-78 (FORTRAN)
     * @author unknown - 25-06-79 UPDATED (FORTRAN)
     * @author Alain Bouvet - converted from fortran to java 04/98
     * @author Peter Braden - Rewritten - 06/07
     * @author Akin Yucelir - review - 07/08
     * @author Aidan Campbell - Converted from java to javascript 06/17
     */
    calculateUBMatrix() {
        try { 
            // intermediate variables
            let cosB2 = (this.gamma.cos * this.alpha.cos - this.beta.cos) / (this.gamma.sin * this.alpha.sin);
            let cosB3 = (this.alpha.cos * this.beta.cos - this.gamma.cos) / (this.alpha.sin * this.beta.sin);
            let sinB2 = Math.sqrt(1 - cosB2 * cosB2);
            let sinB3 = Math.sqrt(1 - cosB3 * cosB3);

            this.b_matrix.fromArray(
                [this.a_rec, this.b_rec * cosB3, this.c_rec * cosB2,
                0, this.b_rec * sinB3, -this.c_rec * sinB2 * this.alpha.cos,
                0, 0, 2* Math.PI / this.c]);
            this.b_matrix.transpose();

            let a = new THREE.Vector3(this.orientation1[0], this.orientation1[1], this.orientation1[2]);
            let b = new THREE.Vector3(this.orientation2[0], this.orientation2[1], this.orientation2[2]);

            a.applyMatrix3(this.b_matrix);
            b.applyMatrix3(this.b_matrix);

            // plane on vector a => calculate vector normal to plane = axb
            let c = a.clone();
            c.cross(b);
            let d = c.clone();
            d.cross(a);

            // transform vectors to unit lengths
            a.normalize();
            c.normalize();
            d.normalize();
            //console.log("a = " + a.getComponent(0) + " " + a.getComponent(1) + " " + a.getComponent(2));
            //console.log("c = " + c.getComponent(0) + " " + c.getComponent(1) + " " + c.getComponent(2));
           // console.log("d = " + d.getComponent(0) + " " + d.getComponent(1) + " " + d.getComponent(2));

            this.u_matrix.fromArray([
                a.getComponent(0), a.getComponent(1), a.getComponent(2),
                d.getComponent(0), d.getComponent(1), d.getComponent(2),
                c.getComponent(0), c.getComponent(1), c.getComponent(2)]);
            this.u_matrix.transpose();

            this.ub_matrix = this.u_matrix.clone();
            this.ub_matrix.multiply(this.b_matrix);

        } catch (e) {
            console.log("Exception in calculateUBMatrix: " + e);
        }

    }
    setA_rec(a_reca) {
        if (a_reca > 0 && this.isReal(a_reca)) {
            this.a_rec = a_reca;
            this.a_rec_valid = true;
        } else {
            this.a_rec_valid = false;
        }
        /*switch (Mode.fromString(type.getMode("a"))) {
            case Mode.ABC:
                if (this.b_rec != a_reca) {
                    this.setB_rec(a_reca);
                }
                if (this.c_rec != a_reca) {
                    setC_rec(a_reca);
                }

                break;
            case Mode.AB:
                if (this.b_rec != a_reca) {
                    this.setB_rec(a_reca);
                }
                break;
            case AC:
                if (this.c_rec != a_reca) {
                    this.setC_rec(a_reca);
                }
                break;
            case Mode.BC:

                break;
            case Mode.NONE:

                break;

        }*/
    }


    setB_rec(b_reca) {
        if (b_reca > 0 && this.isReal(b_reca)) {
            this.b_rec = b_reca;
            this.b_rec_valid = true;
        } else {
            this.b_rec_valid = false;
        }
        /*switch (Mode.fromString(type.getMode("b"))) {
            case Mode.ABC:
                if (this.a_rec != b_reca) {
                    this.setA_rec(b_reca);
                }
                if (this.c_rec != b_reca) {
                    setC_rec(b_reca);
                }

                break;
            case AB:
                if (this.a_rec != b_reca) {
                    setA_rec(b_reca);
                }
                break;
            case AC:
                break;
            case BC:
                if (this.c_rec != b_reca) {
                    setC_rec(b_reca);
                }
                break;
            case NONE:

                break;

        }*/

    }
    setC_rec(c_reca) {
        if (c_reca > 0 && this.isReal(c_reca)) {
            this.c_rec = c_reca;
            this.c_rec_valid = true;
        } else {
            this.c_rec_valid = false;
        }
        /*switch (Mode.fromString(type.getMode("c"))) {
            case Mode.ABC:
                if (this.a_rec != c_reca) {
                    this.setA_rec(c_reca);
                }
                if (this.b_rec != c_reca) {
                    this.setB_rec(c_reca);
                }

                break;
            case Mode.AB:
                break;
            case Mode.AC:
                if (this.a_rec != c_reca) {
                    this.setA_rec(c_reca);
                }
                break;
            case Mode.BC:
                if (this.b_rec != c_reca) {
                    this.setB_rec(c_reca);
                }
                break;
            case Mode.NONE:

                break;

        }*/
    }


    setAlpha_rec(alpha_reca) {
        if (!alpha_reca.isReal() || alpha_reca.greaterThan(Angle.DEG_180) || alpha_reca.lessThan(Angle.ZERO)
            || alpha_reca.equals(Angle.ZERO)) {
            this.alpha_rec_valid = false;
        } else {
            this.alpha_rec = alpha_reca;
            this.alpha_rec_valid = true;
        }
        /*switch (Mode.fromString(type.getMode("alpha"))) {
            case Mode.ABC:
                if (!this.gamma_rec.equals(alpha_reca)) {
                    this.setGamma_rec(alpha_reca);
                }
                if (!this.beta_rec.equals(alpha_reca)) {
                    this.setBeta_rec(alpha_reca);
                }

                break;
            case Mode.AB:
                if (!this.beta_rec.equals(alpha_reca)) {
                    this.setBeta_rec(alpha_reca);
                }
                break;
            case Mode.AC:
                if (!this.gamma_rec.equals(alpha_reca)) {
                    this.setGamma_rec(alpha_reca);
                }
                break;
            case Mode.BC:

                break;
            case Mode.NONE:

                break;
        }*/
    }

    setBeta_rec(beta_reca) {
        if (!beta_reca.isReal() || beta_reca.greaterThan(Angle.DEG_180) || beta_reca.lessThan(Angle.ZERO)
            || beta_reca.equals(Angle.ZERO)) {
            this.beta_rec_valid = false;
        } else {
            this.beta_rec = beta_reca;
            this.beta_rec_valid = true;
        }
        /*switch (Mode.fromString(type.getMode("beta"))) {
            case Mode.ABC:
                if (!this.gamma_rec.equals(beta_reca)) {
                    this.setGamma_rec(beta_reca);
                }
                if (!this.alpha_rec.equals(beta_reca)) {
                    this.setAlpha_rec(beta_reca);
                }

                break;
            case Mode.AB:
                if (!this.alpha_rec.equals(beta_reca)) {
                    this.setAlpha_rec(beta_reca);
                }
                break;
            case Mode.AC:

                break;
            case Mode.BC:
                if (!this.gamma_rec.equals(beta_reca)) {
                    this.setGamma_rec(beta_reca);
                }
                break;
            case Mode.NONE:

                break;

        }*/
    }
    setGamma_rec(gamma_reca) {
        if (!gamma_reca.isReal() || gamma_reca.greaterThan(Angle.DEG_180) || gamma_reca.lessThan(Angle.ZERO)
            || gamma_reca.equals(Angle.ZERO)) {
            this.gamma_rec_valid = false;
        } else {
            this.gamma_rec = gamma_reca;
            this.gamma_rec_valid = true;
        }
        /*switch (Mode.fromString(type.getMode("gamma"))) {
            case Mode.ABC:
                if (!this.beta_rec.equals(gamma_reca)) {
                    this.setBeta_rec(gamma_reca);
                }
                if (!this.alpha_rec.equals(gamma_reca)) {
                    this.setAlpha_rec(gamma_reca);
                }
                break;
            case Mode.AB:
                break;
            case Mode.AC:
                if (!this.alpha_rec.equals(gamma_reca)) {
                    this.setAlpha_rec(gamma_reca);
                }
                break;
            case Mode.BC:
                if (!this.beta_rec.equals(gamma_reca)) {
                    this.setBeta_rec(gamma_reca);
                }
                break;
            case Mode.NONE:

                break;
        }*/
    }

    setVol_rec(vol_rec) {
        if (vol_rec > 0 && this.isReal(vol_rec)) {
            this.vol_rec = vol_rec;
            this.vol_rec_valid = true;
        } else {
            this.vol_rec_valid = false;
        }

    }

    setVol_real(vol_real) {
        if (vol_real > 0 && this.isReal(vol_real)) {
            this.vol_real = vol_real;
            this.vol_real_valid = true;
        } else {
            this.vol_real_valid = false;
        }
    }

    isReal(value) {
        return !(Number.isNaN(value) || !Number.isFinite(value));
    }

    set setA(value){
        this.a = value;
        this.a_valid = true;
    }
    set setB(value){
        this.b = value;
        this.b_valid = true;
    }
    set setC(value){
        this.c = value;
        this.c_valid = true;
    }
    set setAlpha(value){
        this.alpha = value;
        this.alpha_valid = true;
    }
    set setBeta(value){
        this.beta = value;
        this.beta_valid = true;
    }
    set setGamma(value){
        this.gamma = value;
        this.gamma_valid = true;
    }


    set setA_rec_valid(value){
        this.a_rec_valid = value;
    }
    set setB_rec_valid(value){
        this.b_rec_valid = value;
    }
    set setC_rec_valid(value){
        this.c_rec_valid = value;
    }
    set setAlpha_rec_valid(value){
        this.alpha_rec_valid = value;
    }
    set setBeta_rec_valid(value){
        this.beta_rec_valid = value;
    }
    set setGamma_rec_valid(value){
        this.gamma_rec_valid = value;
    }
    set setVol_real(value){
        this.vol_real = value;
        this.vol_real_valid = true;
    }
    set setVol_rec(value){
        this.vol_rec = value;
        this.vol_rec_valid = true;
    }
}

module.exports = Sample;