const THREE = require('three');
const Model = require('./model');
const vEXPController = require('../vexp-controller');
const PubSub = require('pubsub-js');

const FIXED_KI = 1;
const FIXED_KF = 2;

class ScatteringModel extends Model {
    constructor(ki, kf, ss, qh, qk, ql, fx, a1, a2, a3, a3p, a4, a5, a6, status) {
        super();

        this._ki = ki;
        this._kf = kf;
        this._q = new THREE.Vector3(qh, qk, ql).applyMatrix3(vEXPController.bMatrix).length();
        this._qh = qh;
        this._qk = qk;
        this._ql = ql;

        this._kiPoints = {};
        this._kfPoints = {};

        this._kiPoints = {
            start: new THREE.Vector3(),
            end: new THREE.Vector3()
        };

        this._kfPoints = {
            start: new THREE.Vector3(),
            end: new THREE.Vector3()
        };

        this._a1 = a1;
        this._a2 = a2;
        this._a3 = a3;
        this._a3p = a3p;
        this._a4 = a4;
        this._a5 = a5;
        this._a6 = a6;

        this._fixedKi = false;
        this._fixedKf = false;
        this._fixedQ = false;
        this._fixedDE = false;

        // Convert the Nomad fx mode.
        if (fx == FIXED_KI) {
            this._fixedKi = true;
        }
        else if (fx == FIXED_KF) {
            this._fixedKf = true;
        }

        this._ss = ss;
        this._status = status;

        this.updateScattering();

        // Update angles from vEXPController. Due tu calculation response.
        PubSub.subscribe('update-angles', () => { 
            this.setAngles(vEXPController.a1, vEXPController.a2, vEXPController.a3, vEXPController.a3p, vEXPController.a4, vEXPController.a5, vEXPController.a6);
        });
    }

    // Getters.
    get ki() { return this._ki; }
    get kf() { return this._kf; }
    get q() { return this._q; }
    get qh() { return this._qh; }
    get qk() { return this._qk; }
    get ql() { return this._ql; }
    get a1() { return this._a1; }
    get a2() { return this._a2; }
    get a3() { return this._a3; }
    get a3p() { return this._a3p; }
    get a4() { return this._a4; }
    get a5() { return this._a5; }
    get a6() { return this._a6; }
    get kiPoints() { return this._kiPoints; }
    get kfPoints() { return this._kfPoints; }
    get fixedKi() { return this._fixedKi; }
    get fixedKf() { return this._fixedKf; }
    get fixedQ() { return this._fixedQ; }
    get fixedDE() { return this._fixedDE; }

    // Setters.
    set ki(value) {
        this._ki = value; 
        vEXPController.ki = value;
    }

    set kf(value) {
        this._kf = value;
        vEXPController.kf = value;
    }

    set q(value) {
        this._q = value;
        let newQ = new THREE.Vector3(this._qh, this._qk, this._ql).setLength(value).applyMatrix3(vEXPController.bMatrix_1);
        this._qh = newQ.x;
        this._qk = newQ.y;
        this._ql = newQ.z;
        vEXPController.setQFromUI(this._qh, this._qk, this._ql);
    }

    set kiPoints(value) {
        this._kiPoints = value;
        this._ki = value.end.distanceTo(value.start);
        vEXPController.ki = this._ki;
        this.updateScattering();
    }

    set kfPoints(value) {
        this._kfPoints = value;
        this._kf = value.end.distanceTo(value.start);
        vEXPController.kf = this._kf;
        this.updateScattering();
    }

    set fixedKi(value) {
        this._fixedKi = value;

        if (value) {
            vEXPController.fx = FIXED_KI;
        }
    }
    
    set fixedKf(value) {
        this._fixedKf = value;

        if (value) {
            vEXPController.fx = FIXED_KF;
        }
    }
    
    set fixedQ(value) {
        this._fixedQ = value;
    }

    set fixedDE(value) {
        this._fixedDE = value;
    }

    set ss(value) {
        this._ss = value;
        vEXPController.ss = value;
    }
   
    /**
     * Set all Q coordinates.
     */
    setQ(qh, qk, ql) {
        this._qh = qh;
        this._qk = qk;
        this._ql = ql;
        this._q = new THREE.Vector3(this._qh, this._qk, this._ql).applyMatrix3(vEXPController.bMatrix).length();
        vEXPController.setQFromUI(qh, qk, ql);
    }

    /**
     * Set all angles.
     */
    setAngles(a1, a2, a3, a3p, a4, a5, a6) {
        this._a1 = a1;
        this._a2 = a2;
        this._a3 = a3;
        this._a3p = a3p;
        this._a4 = a4;
        this._a5 = a5;
        this._a6 = a6;
    }

    /**
     * Update scattering parameters to recalculate angles
     * when ki, kf and Q value change.
     */
    updateScattering() {
        this.setAngles(vEXPController.a1, vEXPController.a2, vEXPController.a3, vEXPController.a3p, vEXPController.a4, vEXPController.a5, vEXPController.a6);
    }

    /**
     * Cast the model into a basic JS object.
     */
    toObject() {
        let kiStart = this._kiPoints.start;
        let kiEnd = this._kiPoints.end;

        let kfStart = this._kfPoints.start;
        let kfEnd = this._kfPoints.end;

        return {
            ki: this._ki, kf: this._kf, q: this._q,
            qh: this._qh, qk: this._qk, ql: this._ql,

            kiPoints: { start: { x: kiStart.x, y: kiStart.y, z: kiStart.z }, end: { x: kiEnd.x, y: kiEnd.y, z: kiEnd.z } },
            kfPoints: { start: { x: kfStart.x, y: kfStart.y, z: kfStart.z }, end: { x: kfEnd.x, y: kfEnd.y, z: kfEnd.z } },

            a1: this._a1, a2: this._a2, a3: this._a3, a3p: this._a3p, a4: this._a4, a5: this._a5, a6: this._a6,

            fixedKi: this._fixedKi, fixedKf: this._fixedKf, fixedQ: this._fixedQ, fixedDE: this._fixedDE,

            ss: this._ss,

            status: this._status
        };
    }

    /**
     * Save the model data into a config object (the saver object).
     */
    save(saver) {
        saver.scattering = this.toObject();
    }

    /**
     * Restore the model state saved in the saver object (the configuration object).
     */
    restore(saver) {
        let scattering = saver.scattering;

        let kiStart = scattering.kiPoints.start;
        let kiEnd = scattering.kiPoints.end;

        let kfStart = scattering.kfPoints.start;
        let kfEnd = scattering.kfPoints.end;

        let kiPoints = { start: new THREE.Vector3(kiStart.x, kiStart.y, kiStart.z), end: new THREE.Vector3(kiEnd.x, kiEnd.y, kiEnd.z) };
        this._kiPoints = kiPoints;
        this._ki = kiPoints.end.distanceTo(kiPoints.start);
        vEXPController.ki = this._ki;

        let kfPoints = { start: new THREE.Vector3(kfStart.x, kfStart.y, kfStart.z), end: new THREE.Vector3(kfEnd.x, kfEnd.y, kfEnd.z) };
        this._kfPoints = kfPoints;
        this._kf = kfPoints.end.distanceTo(kfPoints.start);
        vEXPController.kf = this._kf;
        
        this._qh = scattering.qh;
        this._qk = scattering.qk;
        this._ql = scattering.ql;
        this._q = new THREE.Vector3(this._qh, this._qk, this._ql).applyMatrix3(vEXPController.bMatrix).length();
        vEXPController.setQFromUI(this._qh, this._qk, this._ql);

        this.setAngles(scattering.a1, scattering.a2, scattering.a3, scattering.a3p, scattering.a4, scattering.a5, scattering.a6);
        this._fixedKi = scattering.fixedKi;
        this._fixedKf = scattering.fixedKf;
        this._fixedQ = scattering.fixedQ;
        this._fixedDE = scattering.fixedDE;
        this._ss = scattering.ss;

        this.updateScattering();
    }
}

module.exports = ScatteringModel;