/**
 * crystal calculations
 * (using fractional units -> apply B matrix afterwards)
 * @author Tobias Weber <tweber@ill.fr>
 * @date Nov-2017
 * @license see "LICENSE" file
 */

//"use strict";

const eps = 1e-5;
const math = require("mathjs");
math.config({epsilon: eps});


// space groups table
const spaceGroups = require("../../../../resources/space-groups");

// scattering lengths table
const scatteringLengths = require("../../../../resources/scattering_lengths");


/**
 * get the symmetry operations of the space group
 */
function getOperations(operationsOrSpacegroup)
{
    if(typeof(operationsOrSpacegroup) === "string")
        return spaceGroups[operationsOrSpacegroup].ops;
    else
        return operationsOrSpacegroup;
}


/**
 * get the symmetric positions according to the space group
 */
function getSymmetricPositions(operationsOrSpacegroup, position)
{
    const operations = getOperations(operationsOrSpacegroup);
    let all_sympos = [];

    for(let i=0; i<operations.length; ++i)
    {
        const op = operations[i];
        const M = math.matrix([
            [ op[0], op[1], op[2] ],
            [ op[3], op[4], op[5] ],
            [ op[6], op[7], op[8] ]]);
        const v = math.matrix([ op[9], op[10], op[11] ]);

        const sympos = math.add(math.multiply(M, position), v);

        let already_occupied = false;
        for(let iSym=0; iSym<all_sympos.length; ++iSym)
        {
            if(math.deepEqual(all_sympos[iSym], sympos))
            {
                already_occupied = true;
                break;
            }
        }

        if(!already_occupied)
            all_sympos.push(sympos.valueOf());
    }

    //console.log(all_sympos);
    return all_sympos;
}


/**
 * get the symmetric positions according to the space group
 * for all elements of the positions array
 */
function getAllSymmetricPositions(operationsOrSpacegroup, positions)
{
    const operations = getOperations(operationsOrSpacegroup);
    let allpos = [];

    for(let i=0; i<positions.length; ++i)
    {
        const pos = positions[i];
        allpos.push(getSymmetricPositions(operations, pos));
    }

    return allpos;
}


/**
 * get the symmetric positions according to the space group
 * for all elements of the positions array
 * additionally restrict the coordinates to the unit cell
 */
function getAllSymmetricPositionsWithinCell(operationsOrSpacegroup, positions)
{
    const rawatompositions = getAllSymmetricPositions(operationsOrSpacegroup, positions);
    let newatompositions = [];

    // iterate distinct atoms
    for(let rawatomidx=0; rawatomidx<rawatompositions.length; ++rawatomidx)
    {
        const rawpositions = rawatompositions[rawatomidx];
        let newpositions = [];

        // iterate all positions for one atom type
        for(let rawidx=0; rawidx<rawpositions.length; ++rawidx)
        {
            const rawpos = rawpositions[rawidx];
            let newpos = [];

            // iterate x,y,z components
            for(let coordidx=0; coordidx<rawpos.length; ++coordidx)
            {
                let comp = rawpos[coordidx];
                comp = math.mod(comp, 1.);
                if(comp < -0.5) comp += math.abs(math.floor(comp));
                if(comp >= 0.5) comp -= math.abs(math.ceil(comp));
                newpos.push(comp);
            }

            //console.log(newpos);
            newpositions.push(newpos);
        }

        newatompositions.push(newpositions);
    }

    return newatompositions;
}


/**
 * get the coherent neutron scattering length of an atom
 */
function getScatteringLength(atom)
{
    let slen = math.complex(0, 0);
    if(scatteringLengths === null)
        return slen;

    const b = scatteringLengths[atom]["Coh b"];
    const ty = typeof(b);
    if(ty === "number")
    {
        slen.re = b;
    }
    else if(ty === "object")
    {
        slen.re = b[0];
        slen.im = b[1];
    }

    return slen;
}


/**
 * the the coherent neutron scattering lengths of all atoms in the array
 */
function getAllScatteringLengths(atoms)
{
    let sl = [];

    for(let i=0; i<atoms.length; ++i)
    {
        const atom = atoms[i];
        sl.push(getScatteringLength(atom));
    }

    return sl;
}


/**
 * calculate the structure factor, ignoring the scattering length
 */
function getStructureFactorSingle(latticeVector, allPositions, sLength)
{
    const c = math.multiply(2*math.pi, math.complex(0, 1));
    let sfact = math.complex(0, 0);

    for(let i=0; i<allPositions.length; ++i)
    {
        const d = math.dot(latticeVector, allPositions[i]);
        const e = math.multiply(c, d);
        sfact = math.add(sfact, math.exp(e));
    }

    sfact = math.multiply(sfact, sLength);
    return sfact;
}


/**
 * calculate the structure factor
 */
function getStructureFactorDirect(latticeVector, scatteringLengths, allPositions)
{
    let sfact = math.complex(0, 0);

    for(let i=0; i<scatteringLengths.length; ++i)
    {
        const sl = scatteringLengths[i];
        const allpos = allPositions[i];
        const sfact_atom = getStructureFactorSingle(latticeVector, allpos, sl);

        sfact = math.add(sfact, sfact_atom);
    }

    return sfact;
}


/**
 * calculate the structure factor
 */
function getStructureFactor(latticeVector, atoms, positions, operationsOrSpacegroup)
{
    const ops = getOperations(operationsOrSpacegroup);
    let sfact = math.complex(0, 0);

    for(let i=0; i<atoms.length; ++i)
    {
        const atom = atoms[i];
        const pos = positions[i];

        const sl = getScatteringLength(atom);

        const allpos = getSymmetricPositions(ops, pos);
        const sfact_atom = getStructureFactorSingle(latticeVector, allpos, sl);

        sfact = math.add(sfact, sfact_atom);
    }

    return sfact;
}


/**
 * test if the Bragg peak at the given lattice vector is absent
 */
function isPeakAbsent(latticeVector, operationsOrSpacegroup)
{
    const c = math.multiply(2*math.pi, math.complex(0, 1));
    const ops = getOperations(operationsOrSpacegroup);

    for(let i=0; i<ops.length; ++i)
    {
        const op = ops[i];
        const M = math.matrix([
            [ op[0], op[3], op[6] ],
            [ op[1], op[4], op[7] ],
            [ op[2], op[5], op[8] ]]);
        const v = math.matrix([ op[9], op[10], op[11] ]);

        if(!math.deepEqual(v, [0, 0, 0]) &&
            math.deepEqual(latticeVector, math.multiply(M, latticeVector)))
        {
            let e = math.exp(math.multiply(c, math.dot(latticeVector, v)));
            if(!math.equal(e, 1.))
                return true;
        }
    }

    return false;
}


/**
 * calculate the intensity from the structure factor
 */
function getPeakIntensity(structureFactor)
{
    return math.multiply(structureFactor,
        math.conj(structureFactor));
}



/**
 * helper function to make a real number more readable
 */
function formatReal(num, precision=6)
{
    if(math.abs(num) < eps) num = 0.;
    return parseFloat(num.toPrecision(precision));
}


/**
 * print structure factor into a string
 */
function formatStructureFactor(structureFactor, precision=6)
{
    let sf_str = "";

    // complex number
    if(typeof(structureFactor.re) !== "undefined")
    {
        let reVal = structureFactor.re;
        let imVal = structureFactor.im;
        if(math.abs(reVal) < eps) reVal = 0.;
        if(math.abs(imVal) < eps) imVal = 0.;

        const sf_r = formatReal(reVal, precision);
        const sf_i = formatReal(math.abs(imVal), precision);
        const sf_i_m = parseFloat(imVal) < 0.;

        sf_str = sf_r;
        if(!math.equal(imVal, 0.))
            sf_str += (sf_i_m ? " - " : " + ") + sf_i + "i";
    }
    else  // real number
    {
        sf_str = formatReal(structureFactor, precision);
    }

    // absolute square
    const peakIntensity = parseFloat(getPeakIntensity(structureFactor));
    const sf2_str = formatReal(peakIntensity, precision);

    return [sf_str, sf2_str];
}


/**
 * find an op in a list of ops and return the index
 * (or -1 if not found)
 */
function _hasSymOp(allops, op)
{
    for(let opidx=0; opidx<allops.length; ++opidx)
    {
        const op1 = allops[opidx];
        if(math.deepEqual(op1, op))
            return opidx;
    }

    return -1;
}


/**
 * find matching space group based on a given set of symops
 */
function findSpaceGroup(cifops)
{
    for(let sgid in spaceGroups)
    {
        const sg = spaceGroups[sgid];
        const sgops = sg.ops;

        // needs same number of symops
        if(sgops.length !== cifops.length)
            continue;

        let possiblematch = true;
        let opindices = [];
        for(let opidx=0; opidx<sgops.length; ++opidx)
        {   
            const op = sgops[opidx];
            const M = math.matrix([
                [ op[0], op[1], op[2],  op[9]  ],
                [ op[3], op[4], op[5],  op[10] ],
                [ op[6], op[7], op[8],  op[11] ],
                [     0,     0,     0,      1] ] );

            const idx = _hasSymOp(cifops, M);
            if(idx === -1)
            {
                possiblematch = false;
                break;
            }

            opindices.push(idx);
        }

        if(possiblematch)
        {
            //console.log("Possible match: " + sg);
            
            // check if all indices are used (to prevent duplicate ops)
            for(let i=0; i<sgops.length; ++i)
            {
                if(opindices.indexOf(i) < 0)
                {
                    possiblematch = false;
                    break;
                }
            }

            if(possiblematch)
                return sg.id;
        }
    }

    console.error("CIF-defined space group is not in the database.");
    return "unknown";
}


/**
 * try to match cifname to a atom name in the scattering lengths database
 */
function getCleanAtomName(_cifname)
{
    let cifname = _cifname.trim().toLowerCase();

    for(let _atomname in scatteringLengths)
    {
        let atomname = _atomname.trim().toLowerCase();
        
        // exact match
        if(cifname === atomname)
            return _atomname;

        // remove extra symbols
        let cifname2 = cifname.replace(/[0-9+-.,]+/g, "");
        if(cifname2 === atomname)
        {
            console.warn("Approximating atom " + _cifname + " with " + _atomname + " as the original is not in the database.");
            return _atomname;
        }
    }

    console.error("CIF-defined atom " + cifname + " is not in the database.");
    return "unknown";
}

// ----------------------------------------------------------------------------


/**
 * exported data and functions
 */
module.exports =
{
    // data
    "spaceGroups": spaceGroups,
    "scatteringLengths": scatteringLengths,


    // functions
    "getSymmetricPositions": getSymmetricPositions,
    "getAllSymmetricPositions": getAllSymmetricPositions,
    "getAllSymmetricPositionsWithinCell": getAllSymmetricPositionsWithinCell,
    "getAllScatteringLengths": getAllScatteringLengths,

    "getStructureFactor": getStructureFactor,
    "getStructureFactorDirect": getStructureFactorDirect,
    "getPeakIntensity": getPeakIntensity,
    "isPeakAbsent": isPeakAbsent,

    // helper functions
    "formatStructureFactor": formatStructureFactor,
    "formatReal": formatReal,

    "findSpaceGroup": findSpaceGroup,
    "getCleanAtomName": getCleanAtomName,
  };
